from .response import JsonResponse, make_json_response
from .decorators import jsonapi
from .testing import FlaskJsonClient
from .formatting import DynamicJSONEncoder, JsonSerializableBase
from .views import MethodView, RestfulView, methodview



from json import JSONEncoder


class DynamicJSONEncoder(JSONEncoder):
    """ JSON encoder for custom classes:

        Uses __json__() method if available to prepare the object.
        Especially useful for SQLAlchemy models
    """

    def default(self, o):
        # Custom JSON-encodeable objects
        if hasattr(o, '__json__'):
            return o.__json__()

        # Default
        return super(DynamicJSONEncoder, self).default(o)


#region SqlAlchemy Tools

try:
    from sqlalchemy import inspect
    from sqlalchemy.orm.state import InstanceState
except ImportError as e:
    def __nomodule(*args, **kwargs): raise e
    inspect = __nomodule
    InstanceState = __nomodule


def get_entity_propnames(entity):
    """ Get entity property names

        :param entity: Entity
        :type entity: sqlalchemy.ext.declarative.api.DeclarativeMeta
        :returns: Set of entity property names
        :rtype: set
    """
    ins = entity if isinstance(entity, InstanceState) else inspect(entity)
    return set(
        ins.mapper.column_attrs.keys() +  # Columns
        ins.mapper.relationships.keys()  # Relationships
    )


def get_entity_loaded_propnames(entity):
    """ Get entity property names that are loaded (e.g. won't produce new queries)

        :param entity: Entity
        :type entity: sqlalchemy.ext.declarative.api.DeclarativeMeta
        :returns: Set of entity property names
        :rtype: set
    """
    ins = inspect(entity)
    keynames = get_entity_propnames(ins)

    # If the entity is not transient -- exclude unloaded keys
    # Transient entities won't load these anyway, so it's safe to include all columns and get defaults
    if not ins.transient:
        keynames -= ins.unloaded

    # If the entity is expired -- reload expired attributes as well
    # Expired attributes are usually unloaded as well!
    if ins.expired:
        keynames |= ins.expired_attributes

    # Finish
    return keynames


class JsonSerializableBase(object):
    """ Declarative Base mixin to allow objects serialization

        Defines interfaces utilized by :cls:ApiJSONEncoder
    """

    def __json__(self, exluded_keys=set()):
        return {name: getattr(self, name)
                for name in get_entity_loaded_propnames(self) - exluded_keys}

#endregion



from flask import current_app, request, Response, json


class JsonResponse(Response):
    """ Response from a JSON API view """

    @staticmethod
    def _lazy_json(data, indent=None): # CHECKME: This hack prevents Flask from providing the correct 'Content-Length' header. is it ok?
        """ Lazy iterable: jsonify() only when iterated """
        yield json.dumps(data, indent=indent)

    def __init__(self, response, status=None, headers=None, **kwargs):
        """ Init a JSON response
        :param response: Response data
        :type response: *
        :param status: Status code
        :type status: int|None
        :param headers: Additional headers
        :type headers: dict|None
        """
        # Store response
        self._response_data = self.preprocess_response_data(response)

        # Init super
        indent = 2 if current_app.config['JSONIFY_PRETTYPRINT_REGULAR'] and not request.is_xhr else None
        super(JsonResponse, self).__init__(
            self._lazy_json(self._response_data, indent=indent),
            headers=headers, status=status, mimetype='application/json',
            direct_passthrough=True, **kwargs)

    def preprocess_response_data(self, response):
        """ Preprocess the response data.

        Override this method to have custom handling of the response

        :param response: Return value from the view function
        :type response: *
        :return: Preprocessed value
        """
        return response

    def get_json(self):
        """ Get the response data object (preprocessed) """
        return self._response_data

    def __getitem__(self, item):
        """ Proxy method to get items from the underlying object """
        return self._response_data[item]


def normalize_response_value(rv):
    """ Normalize the response value into a 3-tuple (rv, status, headers)
        :type rv: tuple|*
        :returns: tuple(rv, status, headers)
        :rtype: tuple(Response|JsonResponse|*, int|None, dict|None)
    """
    status = headers = None
    if isinstance(rv, tuple):
        rv, status, headers = rv + (None,) * (3 - len(rv))
    return rv, status, headers


def make_json_response(rv):
    """ Make JsonResponse
    :param rv: Response: the object to encode, or tuple (response, status, headers)
    :type rv: tuple|*
    :rtype: JsonResponse
    """
    # Tuple of (response, status, headers)
    rv, status, headers = normalize_response_value(rv)

    # JsonResponse
    if isinstance(rv, JsonResponse):
        return rv

    # Data
    return JsonResponse(rv, status, headers)



from functools import wraps, update_wrapper, partial

from .response import normalize_response_value, make_json_response


def jsonapi(f):
    """ Declare the view as a JSON API method

        This converts view return value into a :cls:JsonResponse.

        The following return types are supported:
            - tuple: a tuple of (response, status, headers)
            - any other object is converted to JSON
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        rv = f(*args, **kwargs)
        return make_json_response(rv)
    return wrapper



import flask.json
from flask.testing import FlaskClient

from .response import JsonResponse


class FlaskJsonClient(FlaskClient):
    """ JSON-aware test client """

    def open(self, path, json=None, **kwargs):
        """ Open an URL, optionally posting JSON data
        :param path: URI to request
        :type path: str
        :param json: JSON data to post
        :param method: HTTP Method to use. 'POST' by default if data is provided
        :param data: Custom data to post, if required
        """
        # Prepare request
        if json:
            kwargs['data'] = flask.json.dumps(json)
            kwargs['content_type'] = 'application/json'
            kwargs.setdefault('method', 'POST')

        # Request
        rv = super(FlaskJsonClient, self).open(path, **kwargs)
        ':type rv: flask.Response'

        # Response: JSON?
        if rv.mimetype == 'application/json':
            response = flask.json.loads(rv.get_data())
            return JsonResponse(response, rv.status_code, rv.headers)
        return rv



from collections import defaultdict
from flask.views import View, with_metaclass
from flask import request
from werkzeug.exceptions import MethodNotAllowed


def methodview(methods=(), ifnset=None, ifset=None):
    """ Decorator to mark a method as a view.

    NOTE: This should be a top-level decorator!

    :param methods: List of HTTP verbs it works with
    :type methods: str|Iterable[str]
    :param ifnset: Conditional matching: only if the route param is not set (or is None)
    :type  ifnset: str|Iterable[str]|None
    :param ifset:  Conditional matching: only if the route param is     set (and is not None)
    :type  ifset:  str|Iterable[str]|None
    """
    return _MethodViewInfo(methods, ifnset, ifset).decorator


class _MethodViewInfo(object):
    """ Method view info object """

    def decorator(self, func):
        """ Wrapper function to decorate a function """
        func._methodview = self
        return func

    @classmethod
    def get_info(cls, func):
        """ :rtype: _MethodViewInfo|None """
        try: return func._methodview
        except AttributeError: return None

    def __init__(self, methods=None, ifnset=None, ifset=None):
        if isinstance(methods, basestring):
            methods = (methods,)
        if isinstance(ifnset, basestring):
            ifnset = (ifnset,)
        if isinstance(ifset, basestring):
            ifset = (ifset,)

        #: Method verbs, uppercase
        self.methods = frozenset([m.upper() for m in methods]) if methods else None

        #: Conditional matching: route params that should not be set
        self.ifnset = frozenset(ifnset) if ifnset else None

        # : Conditional matching: route params that should be set
        self.ifset  = frozenset(ifset ) if ifset  else None

    def matches(self, verb, params):
        """ Test if the method matches the provided set of arguments

        :param verb: HTTP verb. Uppercase
        :type verb: str
        :param params: Existing route parameters
        :type params: set
        :returns: Whether this view matches
        :rtype: bool
        """
        return (self.ifset   is None or self.ifset          <= params) and \
               (self.ifnset  is None or self.ifnset.isdisjoint(params)) and \
               (self.methods is None or verb in self.methods)


class MethodViewType(type):
    """ Metaclass that collects methods decorated with @methodview """

    def __new__(cls, name, bases, d):
        rv = type.__new__(cls, name, bases, d)

        # Methods
        view_methods = set(rv.methods or [])
        methods_map = defaultdict(dict)
        for name, func in d.items():
            # Collect methods decorated with method()
            info = _MethodViewInfo.get_info(func)
            if name.startswith('_') or info is None:
                continue
            for method in info.methods:
                methods_map[method][name] = info
                view_methods.add(method)

        # Finish
        rv.methods = tuple(sorted(view_methods))  # ('GET', ... )
        rv.methods_map = dict(methods_map)  # { 'GET': {'get': _MethodViewInfo } }
        return rv


class MethodView(with_metaclass(MethodViewType, View)):
    """ Class-based view that dispatches requests to methods decorated with @methodview """

    def _match_view(self, method, route_params):
        """ Detect a view matching the query

        :param method: HTTP method
        :param route_params: Route parameters dict
        :return: Method
        :rtype: Callable|None
        """
        method = method.upper()
        route_params = frozenset(k for k, v in route_params.items() if v is not None)

        for name, info in self.methods_map[method].items():
            if info.matches(method, route_params):
                return getattr(self, name)
        else:
            return None

    def dispatch_request(self, *args, **kwargs):
        view = self._match_view(request.method, kwargs)
        if view is None:
            raise MethodNotAllowed(description='No view implemented for {}({})'.format(request.method, ', '.join(kwargs.keys())))
        return view(*args, **kwargs)

    @classmethod
    def route_as_view(cls, app, name, rules, *class_args, **class_kwargs):
        """ Register the view with an URL route
        :param app: Flask application
        :type app: flask.Flask
        :param name: Unique view name
        :type name: str
        :param rules: List of route rules to use
        :type rules: Iterable[str|werkzeug.routing.Rule]
        :param class_args: Args to pass to object constructor
        :param class_kwargs: KwArgs to pass to object constructor
        :return: View callable
        :rtype: Callable
        """
        view = super(MethodView, cls).as_view(name, *class_args, **class_kwargs)
        for rule in rules:
            app.add_url_rule(rule, view_func=view)
        return view


class RestfulViewType(MethodViewType):
    """ Metaclass that automatically defines REST methods """
    methods_map = {
        # view-name: (needs-primary-key, http-method)
        # Collection methods
        'list':    (False, 'GET'),
        'create':  (False, 'PUT'),
        # Item methods
        'get':     (True,  'GET'),
        'replace': (True,  'POST'),
        'update':  (True,  'PATCH'),
        'delete':  (True,  'DELETE'),
    }

    def __new__(cls, name, bases, d):
        pk = d.get('primary_key', ())

        # Automatically decorate the existing methods
        found_methods = False
        for view_name, (needs_pk, method) in cls.methods_map.items():
            if view_name in d:
                found_methods = True
                d[view_name] = methodview(
                    method,
                    ifnset=None if needs_pk else pk,
                    ifset=pk if needs_pk else None,
                )(d[view_name])

        if found_methods:
            assert pk, 'Primary key for RestfulView "{}" should be defined'.format(name)
            assert isinstance(pk, (list, tuple)), 'Primary key for RestfulView "{}" should be iterable'.format(name)

        # Proceed
        return super(RestfulViewType, cls).__new__(cls, name, bases, d)


class RestfulView(with_metaclass(RestfulViewType, MethodView)):
    """ Method View that automatically defines the following methods:

        Collection:
            GET /   -> list()
            PUT /   -> create()
        Individual item:
            GET /<pk>    -> get()
            POST /<pk>   -> replace()
            PATCH /<pk>  -> update()
            DELETE /<pk> -> delete()

        You just need to specify PK fields
    """

    #: List of route parameters used as a primary key.
    #: If specified -- then we're working with an individual entry, and if not -- with the whole collection
    #: HTTP methods mostly do not intersect, but 'GET' is special here
    primary_key = ()


__all__ = ('methodview', 'MethodView', 'RestfulView')
